/// <reference lib="esnext.asynciterable" />

export declare type AnyIterable<T> = Iterable<T> | AsyncIterable<T>;

declare type CurriedBatchResult = <T, M extends AnyIterable<T>>(curriedIterable: M) => UnwrapAnyIterable<M>;

declare type CurriedBufferResult = <T, M extends AnyIterable<T>>(curriedIterable: M) => UnwrapAnyIterable_2<M>;

declare type CurriedTakeResult = <T, M extends AnyIterable<T>>(curriedIterable: M) => UnwrapAnyIterable_4<M>;

declare type CurriedTimeResult = <T, M extends AnyIterable<T>>(curriedIterable: M) => UnwrapAnyIterable_5<M>;

declare type FlatMapValue<B> = B | AnyIterable<B> | undefined | null | Promise<B | AnyIterable<B> | undefined | null>;

declare interface IReadable {
    once: any;
    read: any;
}

declare interface ITimeConfig {
    progress?: (delta: [number, number], total: [number, number]) => any;
    total?: (time: [number, number]) => any;
}

declare interface IWritable {
    once: any;
    write: any;
    removeListener: any;
}

export declare type Iterableish<T> = Iterable<T> | Iterator<T> | AsyncIterable<T> | AsyncIterator<T>;

declare type UnArrayAnyIterable<A extends Array<AnyIterable<any>>> = A extends Array<AnyIterable<infer T>> ? T : never;

declare type UnwrapAnyIterable<M extends AnyIterable<any>> = M extends Iterable<infer T> ? Iterable<T[]> : M extends AsyncIterable<infer B> ? AsyncIterable<B[]> : never;

declare type UnwrapAnyIterable_2<M extends AnyIterable<any>> = M extends Iterable<infer T> ? Iterable<T> : M extends AsyncIterable<infer B> ? AsyncIterable<B> : never;

declare type UnwrapAnyIterable_3<M extends AnyIterable<any>> = M extends Iterable<infer T> ? T[] : M extends AsyncIterable<infer B> ? Promise<B[]> : never;

declare type UnwrapAnyIterable_4<M extends AnyIterable<any>> = M extends Iterable<infer T> ? Iterable<T> : M extends AsyncIterable<infer B> ? AsyncIterable<B> : never;

declare type UnwrapAnyIterable_5<M extends AnyIterable<any>> = M extends Iterable<infer T> ? Iterable<T> : M extends AsyncIterable<infer B> ? AsyncIterable<B> : never;

export declare function batch(size: number): CurriedBatchResult;

export declare function batch<T, M extends AnyIterable<T>>(size: number, iterable: M): UnwrapAnyIterable<M>;

export declare function buffer(size: number): CurriedBufferResult;

export declare function buffer<T, M extends AnyIterable<T>>(size: number, iterable: M): UnwrapAnyIterable_2<M>;

export declare function collect<T, M extends AnyIterable<T>>(iterable: M): UnwrapAnyIterable_3<M>;

export declare function concat<I extends Array<Iterable<any>>>(...iterables: I): Iterable<UnArrayAnyIterable<I>>;

export declare function concat<I extends Array<AnyIterable<any>>>(...iterables: I): AsyncIterable<UnArrayAnyIterable<I>>;

export declare function consume<T>(iterable: Iterable<T>): void;

export declare function consume<T>(iterable: AsyncIterable<T>): Promise<void>;

export declare function filter<T, S extends T>(filterFunc: (data: T) => data is S): <A extends T>(curriedIterable: AnyIterable<A>) => AsyncIterableIterator<S>;

export declare function filter<T>(filterFunc: (data: T) => boolean | Promise<boolean>): <A>(curriedIterable: AnyIterable<A>) => AsyncIterableIterator<A>;

export declare function filter<T, S extends T>(filterFunc: (data: T) => data is S, iterable: AnyIterable<T>): AsyncIterableIterator<S>;

export declare function filter<T>(filterFunc: (data: T) => boolean | Promise<boolean>, iterable: AnyIterable<T>): AsyncIterableIterator<T>;

export declare function flatMap<T, B>(func: (data: T) => FlatMapValue<B>): (iterable: AnyIterable<T>) => AsyncIterableIterator<B>;

export declare function flatMap<T, B>(func: (data: T) => FlatMapValue<B>, iterable: AnyIterable<T>): AsyncIterableIterator<B>;

export declare function flatTransform<T, R>(concurrency: number): {
    (func: (data: T) => FlatMapValue<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;
    (func: (data: T) => FlatMapValue<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;
};

export declare function flatTransform<T, R>(concurrency: number, func: (data: T) => FlatMapValue<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;

export declare function flatTransform<T, R>(concurrency: number, func: (data: T) => FlatMapValue<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;

export declare function flatten<B>(iterable: AnyIterable<B | AnyIterable<B>>): AsyncIterableIterator<B>;

export declare function fromStream<T>(stream: IReadable): AsyncIterable<T>;

export declare function getIterator<T>(iterable: Iterable<T> | Iterator<T>): Iterator<T>;

export declare function getIterator<T>(iterable: AsyncIterable<T> | AsyncIterator<T>): AsyncIterator<T>;

export declare function getIterator<T>(iterable: AsyncIterable<T> | Iterable<T>): AsyncIterator<T> | Iterator<T>;

export declare function map<T, B>(func: (data: T) => B | Promise<B>): (iterable: AnyIterable<T>) => AsyncIterableIterator<B>;

export declare function map<T, B>(func: (data: T) => B | Promise<B>, iterable: AnyIterable<T>): AsyncIterableIterator<B>;

export declare function merge<I extends Array<AnyIterable<any>>>(...iterables: I): AsyncIterable<UnArrayAnyIterable<I>>;

export declare function parallelFlatMap<T, R>(concurrency: number): {
    (func: (data: T) => R | Promise<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;
    (func: (data: T) => R | Promise<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;
};

export declare function parallelFlatMap<T, R>(concurrency: number, func: (data: T) => R | Promise<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;

export declare function parallelFlatMap<T, R>(concurrency: number, func: (data: T) => R | Promise<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;

export declare function parallelMap<T, R>(concurrency: number): {
    (func: (data: T) => R | Promise<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;
    (func: (data: T) => R | Promise<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;
};

export declare function parallelMap<T, R>(concurrency: number, func: (data: T) => R | Promise<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;

export declare function parallelMap<T, R>(concurrency: number, func: (data: T) => R | Promise<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;

export declare function parallelMerge<I extends Array<AnyIterable<any>>>(...iterables: I): AsyncIterable<UnArrayAnyIterable<I>>;

export declare function pipeline<T0>(firstFn: () => T0): T0;

export declare function pipeline<T0, T1>(a0: () => T0, a1: (a: T0) => T1): T1;

export declare function pipeline<T0, T1, T2>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2): T2;

export declare function pipeline<T0, T1, T2, T3>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3): T3;

export declare function pipeline<T0, T1, T2, T3, T4>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3, a4: (a: T3) => T4): T4;

export declare function pipeline<T0, T1, T2, T3, T4, T5>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3, a4: (a: T3) => T4, a5: (a: T4) => T5): T5;

export declare function pipeline<T0, T1, T2, T3, T4, T5, T6>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3, a4: (a: T3) => T4, a5: (a: T4) => T5, a6: (a: T5) => T6): T6;

export declare function pipeline<T0, T1, T2, T3, T4, T5, T6, T7>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3, a4: (a: T3) => T4, a5: (a: T4) => T5, a6: (a: T5) => T6, a7: (a: T6) => T7): T7;

export declare function pipeline<T0, T1, T2, T3, T4, T5, T6, T7, T8>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3, a4: (a: T3) => T4, a5: (a: T4) => T5, a6: (a: T5) => T6, a7: (a: T6) => T7, a8: (a: T7) => T8): T8;

export declare function pipeline<T0, T1, T2, T3, T4, T5, T6, T7, T8, T9>(a0: () => T0, a1: (a: T0) => T1, a2: (a: T1) => T2, a3: (a: T2) => T3, a4: (a: T3) => T4, a5: (a: T4) => T5, a6: (a: T5) => T6, a7: (a: T6) => T7, a8: (a: T7) => T8, a9: (a: T8) => T9): T9;

export declare function reduce<T, B>(func: (acc: B, value: T) => B): {
    (start: B): (iterable: AnyIterable<T>) => Promise<B>;
    (start: B, iterable: AnyIterable<T>): Promise<B>;
};

export declare function reduce<T, B>(func: (acc: B, value: T) => B, start: B): (iterable: AnyIterable<T>) => Promise<B>;

export declare function reduce<T, B>(func: (acc: B, value: T) => B, start: B, iterable: AnyIterable<T>): Promise<B>;

export declare function take(count: number): CurriedTakeResult;

export declare function take<T, M extends AnyIterable<T>>(count: number, iterable: M): UnwrapAnyIterable_4<M>;

export declare function tap<T>(func: (data: T) => any): (iterable: AnyIterable<T>) => AsyncIterableIterator<T>;

export declare function tap<T>(func: (data: T) => any, iterable: AnyIterable<T>): AsyncIterableIterator<T>;

export declare function time(config?: ITimeConfig): CurriedTimeResult;

export declare function time<T, M extends AnyIterable<T>>(config: ITimeConfig, iterable: M): UnwrapAnyIterable_5<M>;

export declare function transform<T, R>(concurrency: number): {
    (func: (data: T) => R | Promise<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;
    (func: (data: T) => R | Promise<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;
};

export declare function transform<T, R>(concurrency: number, func: (data: T) => R | Promise<R>): (iterable: AnyIterable<T>) => AsyncIterableIterator<R>;

export declare function transform<T, R>(concurrency: number, func: (data: T) => R | Promise<R>, iterable: AnyIterable<T>): AsyncIterableIterator<R>;

export declare function writeToStream(stream: IWritable): (iterable: AnyIterable<any>) => Promise<void>;

export declare function writeToStream(stream: IWritable, iterable: AnyIterable<any>): Promise<void>;
